/** vmmidi- Unix socket to Alsa MIDI device converter for MIDI in VMware.
 * @author Alexandre DENIS -- contact@alexandredenis.net
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <fcntl.h>
#include <time.h>
#include <setjmp.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <sys/poll.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <alsa/asoundlib.h>

static const char vmmidi_path[] = "/tmp/vmmidi0";
static const char vmmidi_dev[]  = "virtual";
static const int  vmmidi_prio   = -10;

static jmp_buf vmmidi_start;

static inline void vmmidi_disconnected(void)
{
  longjmp(vmmidi_start, 1);
}

static inline void checkrc(int rc, const char msg[])
{
  if(rc == -1)
    {
      perror(msg);
      abort();
    }
}

static void vmmidi_setprio(void)
{
  int rc = setpriority(PRIO_PROCESS,0, vmmidi_prio);
  checkrc(rc, "setpriority");
  if(getuid() != geteuid())
    seteuid(getuid());
  assert(geteuid() != 0);
}

static void vmmidi_write_n(int fd, unsigned char*buff, int l)
{
  int done = 0;
  while(done < l)
    {
      int rc = write(fd, buff+done, l-done);
      if(rc == -1)
	vmmidi_disconnected();
      else
	done += rc;
    }
}

static void vmmidi_read_n(int fd, unsigned char*buff, int l)
{
  int done = 0;
  while(done < l)
    {
      int rc = read(fd, buff+done, l-done);
      if(rc <= 0)
	vmmidi_disconnected();
      else
	done += rc;
    }
}

static void vmmidi_read_message(int fd, unsigned char*buff, int*_l)
{
  vmmidi_read_n(fd, buff, 1);
  unsigned char opcode = buff[0];
  switch(opcode)
    {
      /* opcode + 2 bytes */
    case 0x80 ... 0x8F: /* note off */
    case 0x90 ... 0x9F: /* note on */
    case 0xA0 ... 0xAF: /* after-touch */
    case 0xB0 ... 0xBF: /* midi controler */
    case 0xE0 ... 0xEF: /* pitch bend */
    case 0xF2:          /* track position */
      vmmidi_read_n(fd, buff+1, 2);
      *_l = 3;
      break;
      /* opcode + 1 byte */
    case 0xC0 ... 0xCF: /* program change */
    case 0xD0 ... 0xDF: /* channel after-touch */
    case 0xF3:          /* track number */
      vmmidi_read_n(fd, buff+1, 1);
      *_l = 2;
      break;
      /* opcode + 0 byte */
    case 0xF6:          /* tuning request */
    case 0xF7:          /* end of sysex */
    case 0xF8:          /* midi clock */
    case 0xFA:          /* start */
    case 0xFB:          /* continue */
    case 0xFC:          /* stop */
    case 0xFE:          /* active sensing */
    case 0xFF:          /* reset */
      *_l = 1;
      break;
      /* invalid opcodes */
    case 0xF1:
    case 0xF4:
    case 0xF5:
    case 0xF9:
    case 0xFD:
      *_l = 1;
      fprintf(stderr, "vmmidi: invalid opcode %X\n", (unsigned)opcode);
      break;
      /* variable length */
    case 0xF0:         /* sysex */
      {
	int l = 0;
	do
	  {
	    vmmidi_read_n(fd, buff+l+1, 1);
	    l++;
	  }
	while(buff[l] != 0xF7 && !(buff[l] & 0x80));
      *_l = 1 + l;
      }
      break;
    case 0x00 ... 0x7F: /* status byte */
      fprintf(stderr, "vmmidi: lost opcode.\n");
      *_l = 0;
      break;
    default:
      fprintf(stderr, "vmmidi: unexpected opcode %X\n", (unsigned)opcode);
      *_l = 1;
      break;
    }

}

int main(int argc, char**argv)
{
  int rc = -1;
  struct sockaddr_un unaddr =
    {
      .sun_family = AF_UNIX,
      .sun_path   = { 0 }
    };

  vmmidi_setprio();

  /* setup rawmidi device */
  snd_rawmidi_t*rawmidi_in = NULL, *rawmidi_out = NULL;
  rc = snd_rawmidi_open(&rawmidi_in, &rawmidi_out, vmmidi_dev, 0);
  checkrc(rc, "rawmidi_open");

  /* setup vmmidi socket */
  unlink(vmmidi_path);
  int vmm_sock = socket(PF_UNIX, SOCK_STREAM, 0);
  checkrc(vmm_sock, "socket()");
  strcpy(unaddr.sun_path, vmmidi_path);
  rc = bind(vmm_sock, (struct sockaddr*)&unaddr, sizeof(unaddr));
  checkrc(rc, "bind()");
  listen(vmm_sock, 1);

  /* prepare polling */
  const int n = snd_rawmidi_poll_descriptors_count(rawmidi_in);
  const int nfd = n + 1;
  struct pollfd*p = malloc(nfd * sizeof(struct pollfd));
  snd_rawmidi_poll_descriptors(rawmidi_in, &p[1], 1);
  p[0] = (struct pollfd)
    {
      .fd      = -1,
      .events  = POLLIN | POLLERR,
      .revents = 0
    };

 vmmidi_loop:
  fprintf(stderr, "vmmidi: serial: %s <-> alsa: \"%s\"\n", vmmidi_path, vmmidi_dev);
  /* waiting for vmware */
  struct sockaddr_un vmware_unaddr;
  int vmware_unaddrlen = sizeof(vmware_unaddr);
  int v = accept(vmm_sock, (struct sockaddr*)&vmware_unaddr, &vmware_unaddrlen);
  checkrc(v, "accept");
  p[0].fd = v;
  fprintf(stderr, "vmmidi: connected.\n");

  if(setjmp(vmmidi_start) != 0)
    goto vmmidi_end;

  /* dump */
  while(rc != -1)
    {
      const int buffsize = 1024;
      unsigned char buff[buffsize];

      rc = poll(p, nfd, -1);
      checkrc(rc, "poll()");
#ifdef VERBOSE
      struct timeval t;
      gettimeofday(&t, NULL);
#endif
      if(p[0].revents & POLLIN)
	{
	  int l = 0;
	  vmmidi_read_message(v, buff, &l);
	  if(l == 0)
	    continue;
#ifdef VERBOSE
	  fprintf(stderr, "[%ld:%6ld] socket -> midi (%d) [%X]\n",
		  t.tv_sec, t.tv_usec, l, buff[0]);
#endif
	  rc = snd_rawmidi_write(rawmidi_out, buff, l);
	  checkrc(rc, "rawmidi_write");
	  snd_rawmidi_drain(rawmidi_out);
	}

      if(p[1].revents & POLLIN)
	{
	  int l = snd_rawmidi_read(rawmidi_in, buff, buffsize);
#ifdef VERBOSE
	  fprintf(stderr, "[%ld:%6ld} midi -> socket (%d) [%X]\n",
		  t.tv_sec, t.tv_usec, l, buff[0]);
#endif
	  checkrc(l, "ramidi_read");
	  vmmidi_write_n(v, buff, l);
	}
    }
 vmmidi_end:
  fprintf(stderr, "vmmidi: disconnected.\n");
  close(v);
  v = -1;
  goto vmmidi_loop;
  
  return 0;
}
